// VideoTextureManager.cpp

#include "VideoTextureManager.h"
#include "VideoOps.h"
#include "../Strings/StringOps.h"
#include "../Logging/Logger.h"
#include "VideoTexture.h"
#include "../Graphics/DynamicTexture.h"
#include "../FileSystem/FileSystem.h"

using namespace Convex;

Texture *VideoTextureManager::CreateVideoTexture(const UniString &filename)
{
	// try to find it first
	if (video_textures[filename].tex) 
	{
		video_textures[filename].ref_count ++;
		return video_textures[filename].tex->GetTexture();
	}

	// create it
	VideoTexture *ret = VideoOps::CreateVideoTexture(gc, filename, true, &texture_path);
	if (!ret) return NULL;

	video_textures[filename].tex = ret;
	video_textures[filename].ref_count = 1;

	// prepare the initial frame
	ret->PrepareNextFrame();
	ret->UpdatePixels();

	// begin preparing the next frame
	ret->AsyncPrepareNextFrame();

	// start frame clock
	video_textures[filename].clock.Reset();
	video_textures[filename].clock.Start();

	return ret->GetTexture();
}

void VideoTextureManager::DeleteVideoTexture(Texture *tex)
{
	if (!tex) return;
	for (map<UniString, RefVideoTexture>::iterator it=video_textures.begin(); 
		it!=video_textures.end(); it++)
	{
		if (it->second.tex && it->second.tex->GetTexture() == tex)
		{
			it->second.ref_count --;
			if (it->second.ref_count == 0)
			{
				VideoOps::DeleteVideoTexture(it->second.tex);
				video_textures.erase(it);
			}

			break;
		}
	}
}

UniStringList VideoTextureManager::CollectVideoTextures() const
{
	UniStringList collected_textures;

	for (map<UniString, RefVideoTexture>::const_iterator it=video_textures.begin(); 
		it!=video_textures.end(); it++)
	{
		UniString texture_name = it->first;

		UniString full_filename;
		bool found = ConvexFS->FileExists(texture_name, &texture_path, &full_filename);
		if (found ) {
			ConvexInfo("Collecting Texture: " + full_filename);
			collected_textures.push_back(full_filename);
		}
	}

	return collected_textures;
}

VideoTextureManager::VideoTextureManager(GraphicsContext *context) : TextureManager(context)
{

}

VideoTextureManager::~VideoTextureManager()
{
	for (map<UniString, RefVideoTexture>::iterator it=video_textures.begin(); 
		it != video_textures.end(); it++)
	{
		if (it->second.tex) VideoOps::DeleteVideoTexture(it->second.tex);
	}
	video_textures.clear();
}

Texture *VideoTextureManager::CreateTexture(const UniString &filename, int quality)
{
	// see if the file is mpg
	string extension = StringOps::GetExtension(filename.UTF8());
	if (extension == "mpg") return CreateVideoTexture(filename);
	return TextureManager::CreateTexture(filename, quality);
}
void VideoTextureManager::DeleteTexture(Texture *tex)
{
	DeleteVideoTexture(tex);
	TextureManager::DeleteTexture(tex);
}

#ifdef DEVELOPMENT_VERSION

UniStringList VideoTextureManager::CollectTextures() const
{
	UniStringList ret = CollectVideoTextures();
	size_t sz = ret.size();
	UniStringList normal_textures = TextureManager::CollectTextures();
	ret.insert(ret.end(), normal_textures.begin(), normal_textures.end());

	// TODO: Remove this
	assert(sz + normal_textures.size() == ret.size());

	return ret;
}

#endif 

void VideoTextureManager::UpdateVideoTextures()
{
	for (map<UniString, RefVideoTexture>::iterator it = video_textures.begin();
		it != video_textures.end(); it++)
	{
		if (!it->second.tex) continue;
		if (it->second.clock.GetSecs() >= it->second.tex->GetFrameDuration())
		{
			// update
			it->second.clock.Reset();
			it->second.tex->Flush();
			it->second.tex->UpdatePixels();

			// loop
			if (it->second.tex->Ended()) 
			{
				it->second.tex->Restart();
			}

			it->second.tex->AsyncPrepareNextFrame();
		}
	}
}